#lang scheme
(require "cub-exp.ss"
         (prefix-in e-e: "../eval/eval-exp.ss")
         redex/reduction-semantics)

(provide exp-red cub-exp+ec subst)

(define-extended-language cub-exp+ec
  cub-exp
  (E hole
     (biop E exp)
     (biop v E)
     (new id v ... E exp ...)
     (new-array id E e)
     (new-array id v E)
     (let ([var id E] dec ...) exp)
     (begin E exp exp ...)
     (if E exp exp)
     lval-E
     (:= lval-E exp)
     (:= lval-v E)
     (loop E))
  (lval-v id 
          (aref id num) 
          (dot id num))
  (lval-E (aref E exp)
          (aref v E)
          (dot E exp)
          (dot v E))  
  (exp ....
       (loop exp))
  (S (fr ...))
  (fr (id v)
      (h:id (record v ...))
      (h:id (array v ...)))
  (v num str nil () h:id)
  (bool #f #t)
  (h:id (side-condition variable_1 
                        (regexp-match #rx"^h:" (symbol->string (term variable_1))))))
  
(define exp-red
  (reduction-relation 
   cub-exp+ec
   (--> (S (in-hole E (biop v_1 v_2)))
        (S (in-hole E (δ biop v_1 v_2)))
        δ)
   (--> (S (in-hole E (begin v exp)))
        (S (in-hole E exp))
        begin2)
   (--> (S (in-hole E (begin v exp_1 exp_2 exp_3 ...)))
        (S (in-hole E (begin exp_1 exp_2 exp_3 ...)))
        beginN)
   
   (--> ((fr ...) (in-hole E (let ([var id_old v_1] dec ...) exp_2)))
        (((id_new v_1) fr ...) (in-hole E (subst id_old id_new (let (dec ...) exp_2))))
        (fresh id_new)
        letN)
   (--> (S (in-hole E (let () exp)))
        (S (in-hole E exp))
        let0)
   (--> (S (in-hole E (let ([type id ty] dec ...) exp)))
        (S (in-hole E (let (dec ...) exp)))
        let-ty)
   (--> (S (in-hole E (let ([var id id_ty exp_r] dec ...) exp)))
        (S (in-hole E (let ([var id exp_r] dec ...) exp)))
        let-idty)
   
   (--> ((fr_before ... (id v) fr_after ...)
         (in-hole E id))
        ((fr_before ... (id v) fr_after ...)
         (in-hole E v))
        get)
   (--> ((fr_before ... (id v_old) fr_after ...)
         (in-hole E (:= id v_new)))
        ((fr_before ... (id v_new) fr_after ...)
         (in-hole E ()))
        set)
   
   (--> (S (in-hole E (if 0 exp_1 exp_2)))
        (S (in-hole E exp_2))
        if0)
   (--> (S (in-hole E (if num exp_1 exp_2)))
        (S (in-hole E exp_1))
        (side-condition (term (non-zero? num)))
        ifN)
   
   (--> (S (in-hole E (when exp_1 exp_2)))
        (S (in-hole E (if exp_1 exp_2 ())))
        when)
   
   (--> (S (in-hole E (for (id exp_1 exp_2) exp_3)))
        (S (in-hole E (let ([var id exp_1]
                            [var id_top exp_2])
                        (when (< id id_top)
                          (begin 
                            exp_3
                            (while (< id id_top)
                                   (begin
                                     (:= id (+ id 1))
                                     exp_3)))))))
        (fresh id_top)
        for)
   
   (--> (S (in-hole E (while exp_1 exp_2)))
        (S (in-hole E (if exp_1 (loop (begin exp_2 (while exp_1 exp_2))) ())))
        while)
   
   (--> (S (in-hole E_1 (loop (in-hole E_2 (break)))))
        (S (in-hole E_1 ()))
        break
        (side-condition (term (no-loop (in-hole E_2 (break))))))
   (--> (S (in-hole E (loop v)))
        (S (in-hole E v))
        loop)
   
   (--> ((fr ...) 
         (in-hole E (new id v ...)))
        (((h:id (record v ...)) fr ...)
         (in-hole E h:id))
        (fresh h:id)
        new)
   
   (--> ((fr_before ... (h:id (record v_1 ... v v_2 ...)) fr_after ...)
         (in-hole E (dot h:id num)))
        ((fr_before ... (h:id (record v_1 ... v v_2 ...)) fr_after ...) 
         (in-hole E v))
        (side-condition (= (term num) (length (term (v_1 ...)))))
        dot)
   
   (--> ((fr_before ... (h:id (record v_1 ... v_old v_2 ...)) fr_after ...)
         (in-hole E (:= (dot h:id num) v_new)))
        ((fr_before ... (h:id (record v_1 ... v_new v_2 ...)) fr_after ...) 
         (in-hole E ()))
        (side-condition (= (term num) (length (term (v_1 ...)))))
        dot-set)
   
   (--> ((fr ...) 
         (in-hole E (new-array id num v)))
        (((h:id (array ,@(term (n-of num v))) fr ...))
         (in-hole E h:id))
        (fresh h:id)
        new-array)
   
   (--> ((fr_before ... (h:id (array v_1 ... v v_2 ...)) fr_after ...)
         (in-hole E (aref h:id num)))
        ((fr_before ... (h:id (array v_1 ... v v_2 ...)) fr_after ...) 
         (in-hole E v))
        (side-condition (= (term num) (length (term (v_1 ...)))))
        aref)
   
   (--> ((fr_before ... (h:id (array v_1 ... v_old v_2 ...)) fr_after ...)
         (in-hole E (:= (aref h:id num) v_new)))
        ((fr_before ... (h:id (array v_1 ... v_new v_2 ...)) fr_after ...) 
         (in-hole E ()))
        (side-condition (= (term num) (length (term (v_1 ...)))))
        aset)))

;; a hack to avoid the multiple match problem.
(define loop? (redex-match cub-exp+ec (in-hole E (loop exp))))
(define-metafunction cub-exp+ec
  no-loop : exp -> bool
  [(no-loop exp) ,(not (loop? (term exp)))])

(define-metafunction cub-exp+ec
  δ : biop v v -> v
  [(δ biop v_1 v_2) ,(e-e:δ (term biop) (term v_1) (term v_2))])

(define-metafunction cub-exp+ec 
  non-zero? : num -> bool
  [(non-zero? 0) #f]
  [(non-zero? num) #t])

(define-metafunction cub-exp+ec 
  n-of : num v -> (v ...)
  [(n-of num v) ,(build-list (term num) (λ (x) (term v)))])

  
(define-metafunction cub-exp+ec
  subst : id id any -> any
  [(subst id_old id_new id_old) id_new]
  [(subst id_old id_new (let (dec_1 ... [var id_old exp_1] dec_2 ...) exp_2))
   (let ((subst-dec id_old id_new dec_1) ...
         [var id_old (subst id_old id_new exp_1)]
         dec_2 ...)
     exp_2)]
  [(subst id_old id_new (let (dec_1 ...) exp_1))
   (let ((subst-dec id_old id_new dec_1) ...) (subst id_old id_new exp_1))]
  [(subst id_old id_new (any_1 ...)) ((subst id_old id_new any_1) ...)]
  [(subst id_old id_new any_1) any_1])

(define-metafunction cub-exp+ec
  subst-dec : id id dec -> dec
  [(subst-dec id_old id_new [var id exp])
   [var id (subst id_old id_new exp)]]
  [(subst-dec id_old id_new [var id id_ty exp])
   [var id id_ty (subst id_old id_new exp)]]
  [(subst-dec id_old id_new [type id ty])
   [type id ty]])
