#lang scheme
(require "exp-red.ss"
         redex/reduction-semantics)

(test-equal (term (subst x y x))
            (term y))
(test-equal (term (subst x y z))
            (term z))
(test-equal (term (subst x y (+ x (+ y z))))
            (term (+ y (+ y z))))
(test-equal (term (subst x y (let ([var z x]) (+ z x))))
            (term (let ([var z y]) (+ z y))))

(test-equal (term (subst x y (let ([var p x]) (+ p x))))
            (term (let ([var p y]) (+ p y))))
(test-equal (term (subst x y (let ([var p x] [var q x]) (+ p x))))
            (term (let ([var p y] [var q y]) (+ p y))))
(test-equal (term (subst x y (let ([var p x] [var x x] [var w x]) (+ p x))))
            (term (let ([var p y] [var x y] [var w x]) (+ p x))))

(test-->> exp-red 
          (term (() (+ 1 2)))
          (term (() 3)))
(test-->> exp-red 
          (term (() (+ (* 3 2) 1)))
          (term (() 7)))

(test-->> exp-red
          (term (() (begin 1 2)))
          (term (() 2)))
(test-->> exp-red
          (term (() (begin 1 2 3 4)))
          (term (() 4)))
(test-->> exp-red
          (term (() (let ([var x 1]) x)))
          (term (((id_new 1)) 1)))
(test-->> exp-red
          (term (() (let ([var x 1][var y x]) y)))
          (term (((id_new1 1) (id_new 1)) 1)))
(test-->> exp-red
          (term (() (let ([var x int 1]) x)))
          (term (((id_new 1)) 1)))
(test-->> exp-red
          (term (() (let ([var x int 1][var y int x]) y)))
          (term (((id_new1 1) (id_new 1)) 1)))

(test-->> exp-red
          (term (() (let ([var x 1]) (:= x 2))))
          (term (((id_new 2)) ())))
(test-->> exp-red
          (term (() (let ([var x 1]) (+ x 1))))
          (term (((id_new 1)) 2)))
(test-->> exp-red
          (term (() (let ([var x 1]) (:= x (+ x 1)))))
          (term (((id_new 2)) ())))
(test-->> exp-red
          (term (() (+ 1 (let ([var x 1]) x))))
          (term (((id_new 1)) 2)))


(test-->> exp-red (term (() (if 0 1 2))) (term (() 2)))
(test-->> exp-red (term (() (if 2 1 0))) (term (() 1)))
(test-->> exp-red 
          (term (() (let ([var x 10])
                      (while x
                             (:= x (- x 1))))))
          (term (((id_new 0)) ())))
(test-->> exp-red 
          (term (() (while 1
                           (break))))
          (term (() ())))

(test-->> exp-red 
          (term (() 
                 (let ([var x 1])
                   (while 1
                          (begin
                            (while 1
                                   (break))
                            (:= x 2)
                            (break))))))
          (term (((id_new 2)) ())))

(test-->> exp-red 
          (term (() 
                 (let ([var x 1])
                   (while x
                          (begin 
                            (when (= x 3)
                              (break))
                            (:= x (+ x 1)))))))
          (term (((id_new 3)) ())))

(test-->> exp-red
          (term (() (let ([var y 0])
                      (begin
                        (for (x 0 5)
                          (:= y (+ y x)))
                        y))))
          '(((id_new2 5) (id_new1 5) (id_new 15)) 15))

(test-->> exp-red
          (term (() (let ([type t (array int)]
                          [var x t (new-array t 4 0)])
                      (aref x 1))))
          (term (((id_new h:id) (h:id (array 0 0 0 0))) 0)))

(test-->> exp-red
          (term (() (let ([type t (array int)]
                          [var x t (new-array t 4 0)])
                      (begin (:= (aref x 1) 2)
                             (aref x 1)))))
          (term (((id_new h:id) (h:id (array 0 2 0 0))) 2)))

(test-->> exp-red
          (term (() (let ([type t (record int string)]
                          [var x t (new t 1 "2")])
                      (dot x 0))))
          (term (((id_new h:id) (h:id (record 1 "2"))) 1)))

(test-->> exp-red
          (term (() (let ([type t (record int string)]
                          [var x t (new t 1 "2")])
                      (begin (:= (dot x 0) 3)
                             (dot x 0)))))
          (term (((id_new h:id) (h:id (record 3 "2"))) 3)))

(test-->> exp-red
          (term (() (let ([type t (record int string)]
                          [var x t (new t 1 "2")])
                      (begin (:= (dot x 1) "3")
                             (dot x 1)))))
          (term (((id_new h:id) (h:id (record 1 "3"))) "3")))
(test-->> exp-red 
          (term (()
                 (let ([var x 2])
                   (begin (when (<= x 2)
                            (:= x (* x x)))
                          x))))
          (term (((id_new 4)) 4)))
(test-->> exp-red 
          (term (()
                 (let ([var x 3])
                   (begin (when (<= x 2)
                            (:= x (* x x)))
                          x))))
          (term (((id_new 3)) 3)))

(test-results)
